{% extends "layout.html" %}

{% block content %}
<div class="container">
  <div class="center row">
    <div class="columns large-4 large-offset-4 medium-6 medium-offset-3">
      <form method="POST" action="">
        {% apply field(name="username", label=_("Username")) %}{% end %}
        {% apply field(name="password", label=_("Password"), type="password") %}{% end %}
        {% apply field(name="server", label=_("Server"), list="serverlist") %}{% end %}
        <datalist id="serverlist">
          {% for opt in sorted(servers().keys()) %}
          <option value="{{opt}}"></option>
          {% end %}
        </datalist>
        <div class="columns large-12 text-center">
          <input type="submit" name="login" class="button" id="Login" value="Login"/>
        </div>
      </form>
    </div>
  </div>
</div>
{% end %}
