{% extends "layout.html" %}
{% block content %}
<div class="dashboard">
    <div class="widgets">

    </div>
    <script type="text/dashboard">
        {% raw to_json(dashboard) %}
    </script>

    <script type="text/datasources">
        {% raw to_json(datasources) %}
    </script>

    <script type="text/datasource_timeline">
        {% raw to_json(timeline) %}
    </script>
</div>
{% end %}
